package filegateway

import (
	"context"
	"fmt"
	"io"
	"os"
	"path"
	"path/filepath"
	"time"

	xbus "xbus.io/go-xbus/v3"
)

func localFilesystemFactory(options FilesystemOptions) (Filesystem, error) {
	if options.Host != "" {
		return nil, fmt.Errorf(
			"Invalid host for a file:// uri. Expects a empty host, got %s", options.Host)
	}
	return &LocalFilesystem{
		prefix:          options.Custom["prefix"].(string),
		pollingInterval: options.PollingInterval,
	}, nil
}

func init() {
	RegisterFilesystem("file", localFilesystemFactory)
}

// LocalFilesystem stores files on the local filesystem
type LocalFilesystem struct {
	prefix          string
	pollingInterval time.Duration
}

func (fs *LocalFilesystem) String() string {
	return "file:///" + fs.prefix
}

// Write ...
func (fs *LocalFilesystem) Write(filepath string, size int) (io.WriteCloser, error) {
	abspath := path.Join(fs.prefix, filepath)
	dirname := path.Dir(abspath)
	if stat, err := os.Stat(dirname); os.IsNotExist(err) {
		err = os.MkdirAll(dirname, 0755)
		if err != nil {
			return nil, err
		}
	} else if !stat.IsDir() {
		return nil, fmt.Errorf(
			"%s is not a directory, cannot write to %s",
			dirname, abspath)
	}
	return os.OpenFile(abspath, os.O_RDWR|os.O_CREATE, 0755)
}

// Observe naïvely do a file list every second
func (fs *LocalFilesystem) Observe(
	ctx context.Context, pattern string, log xbus.Logger,
) (<-chan string, error) {
	fullpattern := pattern
	if fs.prefix != "" {
		fullpattern = path.Join(fs.prefix, pattern)
	}
	c := make(chan string)
	fo := NewFileObserver(ctx, pattern, fs.prefix, c)
	_, err := filepath.Glob(fullpattern)
	if err != nil {
		return nil, err
	}
	go func() {
		for {
			matches, err := filepath.Glob(fullpattern)
			if err != nil {
				log.Errorf("file: Error listing files: %s", err)
				return
			}
			if err := fo.AddFiles(matches...); err != nil {
				log.Errorf("file: Error emitting files: %s", err)
				return
			}
			select {
			case <-time.After(fs.pollingInterval):
			case <-ctx.Done():
				return
			}
		}
	}()
	return c, nil
}

// Read returns an opened File
func (fs *LocalFilesystem) Read(filepath string) (io.ReadCloser, error) {
	return os.Open(path.Join(fs.prefix, filepath))
}

// WaitForDeletion wath the given path every second until it does not exist
// anymore
func (fs *LocalFilesystem) WaitForDeletion(ctx context.Context, filepath string) error {
	abspath := path.Join(fs.prefix, filepath)
	for {
		if _, err := os.Stat(abspath); err != nil {
			if os.IsNotExist(err) {
				return nil
			}
			return err
		}
		select {
		case <-time.After(fs.pollingInterval):
		case <-ctx.Done():
			return ctx.Err()
		}
	}
}

// Rename renames a file
func (fs *LocalFilesystem) Rename(oldpath, newpath string) error {
	return os.Rename(
		path.Join(fs.prefix, oldpath),
		path.Join(fs.prefix, newpath),
	)
}

// Delete removes a file
func (fs *LocalFilesystem) Delete(filepath string) error {
	return os.Remove(path.Join(fs.prefix, filepath))
}

func init() {
	MustRegisterFilesystem("file", localFilesystemFactory)
}
